from enum import Enum
from typing import Any

from pydantic import BaseModel


class FunctionDefinitionFormat(str, Enum):
    BASIC = "basic"  # name and description only
    OPENAI = "openai"  # openai function call format
    ANTHROPIC = "anthropic"  # anthropic function call format


class GetFunctionDefinitionParams(BaseModel):
    """Parameters for getting a function definition.

    The backend requires "format" parameter but this value should be set by the
    developer and not LLM when using the sdk, so the 'format' parameter is not present in meta SCHEMA.
    """

    function_name: str
    format: FunctionDefinitionFormat


class FunctionExecutionParams(BaseModel):
    """Parameters for executing a function.

    The function requires two key parameters:
    1. function_name: The name of the function to execute, which is the function name of the function that is
    retrieved using the ACI_GET_FUNCTION_DEFINITION meta function.
    2. function_arguments: A dictionary containing all input arguments required to execute
    the specified function. These arguments are also provided by the function definition
    retrieved using the ACI_GET_FUNCTION_DEFINITION meta function. If a function does not require input arguments, an empty dictionary should be provided.
    3. linked_account_owner_id: to specify with credentials of which linked account the
    function should be executed.
    """

    function_name: str
    function_arguments: dict
    linked_account_owner_id: str


class FunctionExecutionResult(BaseModel):
    """Result of a Aipolabs ACI indexed function (e.g. "BRAVE_SEARCH__WEB_SEARCH") execution.
    Should be identical to the class defined on server side.
    """

    success: bool
    data: Any | None = None
    error: str | None = None


class SearchFunctionsParams(BaseModel):
    """Parameters for searching functions.

    Parameters should be identical to the ones on the server side.
    """

    app_names: list[str] | None = None
    intent: str | None = None
    allowed_apps_only: bool = False
    format: FunctionDefinitionFormat = FunctionDefinitionFormat.OPENAI
    limit: int | None = None
    offset: int | None = None


class Function(BaseModel):

    name: str
    description: str


class FunctionDetails(BaseModel):
    id: str
    app_name: str
    name: str
    description: str
    tags: list[str]
    visibility: str
    active: bool
    protocol: str
    protocol_data: dict
    parameters: dict
    response: dict



from pydantic import BaseModel

from aipolabs.types.functions import Function, FunctionDetails


class SearchAppsParams(BaseModel):
    """Parameters for filtering applications.

    Parameters should be identical to the ones on the server side.

    """

    intent: str | None = None
    allowed_apps_only: bool = False
    include_functions: bool = False
    categories: list[str] | None = None
    limit: int | None = None
    offset: int | None = None


class AppBasic(BaseModel):
    """Representation of an application. Search results will return a list of these."""

    # instance attributes should match the schema defined on the server side.
    name: str
    description: str
    functions: list[Function] | None = None


class AppDetails(BaseModel):
    """Detailed representation of an application, returned by App.get()."""

    id: str
    name: str
    display_name: str
    provider: str
    version: str
    description: str
    logo: str | None
    categories: list[str]
    visibility: str
    active: bool
    security_schemes: list[str]
    functions: list[FunctionDetails]



"""
This module defines the ACI_EXECUTE_FUNCTION meta function, which is used by LLM to submit
execution requests for indexed functions on aipolabs ACI backend.

This module includes the schema definition for the function and a Pydantic model for
validating the execution parameters.
"""

import aipolabs.meta_functions._aci_get_function_definition as ACIGetFunctionDefinition

NAME = "ACI_EXECUTE_FUNCTION"
SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "Execute a specific retrieved function. Provide the executable function name, and the required function parameters for that function based on function definition retrieved.",
        "parameters": {
            "type": "object",
            "properties": {
                "function_name": {
                    "type": "string",
                    "description": f"The name of the function to execute, which is retrieved from the {ACIGetFunctionDefinition.NAME} function.",
                },
                "function_arguments": {
                    "type": "object",
                    "description": "A dictionary containing key-value pairs of input parameters required by the specified function. The parameter names and types must match those defined in the function definition previously retrieved. If the function requires no parameters, provide an empty object.",
                    "additionalProperties": True,
                },
            },
            "required": ["function_name", "function_arguments"],
            "additionalProperties": False,
        },
    },
}


def wrap_function_arguments_if_not_present(obj: dict) -> dict:
    if "function_arguments" not in obj:
        # Create a copy of the input dict
        processed_obj = obj.copy()
        if "function_name" not in processed_obj:
            raise ValueError("function_name is required")
        # Extract function_name
        function_name = processed_obj.pop("function_name")
        # Create new dict with correct structure
        processed_obj = {
            "function_name": function_name,
            "function_arguments": processed_obj,  # All remaining fields go here
        }
        return processed_obj
    return obj



"""
This module defines the ACI_SEARCH_FUNCTIONS_WITH_INTENT meta function, which is used by LLM to search for
relevant executable functions that can help complete a task.

This function is similar to ACI_SEARCH_FUNCTIONS, but it removes the app_names parameter for a
simplified workflow (Otherwise you'd have to call ACI_SEARCH_APPS first to get the app names).
"""

NAME = "ACI_SEARCH_FUNCTIONS_WITH_INTENT"
SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to find relevant executable functions (tools) that can help complete your tasks or get data and information you need.",
        "parameters": {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Use this to discover relevant functions (tools) you might need. Returned results of this function will be sorted by relevance to the intent. Examples include 'what's the top news in the stock market today', 'i want to automate outbound marketing emails'.",
                },
                "limit": {
                    "type": "integer",
                    "default": 100,
                    "description": "The maximum number of functions to return from the search per response.",
                    "minimum": 1,
                    "maximum": 1000,
                },
                "offset": {
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                    "description": "Pagination offset.",
                },
            },
            "required": [],
            "additionalProperties": False,
        },
    },
}



import aipolabs.meta_functions._aci_execute_function as ACIExecuteFunction
import aipolabs.meta_functions._aci_get_function_definition as ACIGetFunctionDefinition
import aipolabs.meta_functions._aci_search_apps as ACISearchApps
import aipolabs.meta_functions._aci_search_functions as ACISearchFunctions
import aipolabs.meta_functions._aci_search_functions_with_intent as ACISearchFunctionsWithIntent

__all__ = [
    "ACISearchApps",
    "ACISearchFunctions",
    "ACISearchFunctionsWithIntent",
    "ACIGetFunctionDefinition",
    "ACIExecuteFunction",
]



"""
This module defines the ACI_GET_FUNCTION_DEFINITION meta function, which is used by LLM to retrieve
the definition of an executable function.

The function requires one key parameter:
1. function_name: The name of the function to get the definition for. e.g. `BRAVE_SEARCH__WEB_SEARCH`

It returns json schema of the function definition.
"""

NAME = "ACI_GET_FUNCTION_DEFINITION"

SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to get the definition of an executable function.",
        "parameters": {
            "type": "object",
            "properties": {
                "function_name": {
                    "type": "string",
                    "description": "The name of the function you want to get the definition for. You can get function names by using the ACI_SEARCH_FUNCTIONS function.",
                }
            },
            "required": ["function_name"],
            "additionalProperties": False,
        },
    },
}



class AipolabsACIError(Exception):
    """Base exception for all Aipolabs SDK errors"""

    def __init__(self, message: str):
        super().__init__(message)


class APIKeyNotFound(AipolabsACIError):
    """Raised when the API key is not found."""

    pass


class AuthenticationError(AipolabsACIError):
    """Raised when there are authentication issues (401)"""

    pass


class PermissionError(AipolabsACIError):
    """Raised when the user doesn't have permission (403)"""

    pass


class NotFoundError(AipolabsACIError):
    """Raised when the requested resource is not found (404)"""

    pass


class ValidationError(AipolabsACIError):
    """Raised when the request is invalid (400)"""

    pass


class RateLimitError(AipolabsACIError):
    """Raised when rate limit is exceeded (429)"""

    pass


class ServerError(AipolabsACIError):
    """Raised when server errors occur (500-series)"""

    pass


class UnknownError(AipolabsACIError):
    """Raised when an unknown error occurs"""

    pass



"""
This module defines the ACI_SEARCH_FUNCTIONS meta function, which is used by LLM to search for
relevant executable functions that can help complete a task.

You can filter by adding app names, which can be retrieved using the ACI_SEARCH_APPS meta function.
"""

NAME = "ACI_SEARCH_FUNCTIONS"
SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to find relevant executable functions that can help complete your tasks or get data and information you need.",
        "parameters": {
            "type": "object",
            "properties": {
                "app_names": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The names of the apps you want to search functions for. If provided, the search will be limited to the functions of the specified apps. Use null to search functions across all apps. You can find app names by first using the ACI_SEARCH_APPS function.",
                },
                "intent": {
                    "type": "string",
                    "description": "Use this to find relevant functions you might need. Returned results of this function will be sorted by relevance to the intent. Examples include 'what's the top news in the stock market today', 'i want to automate outbound marketing emails'.",
                },
                "limit": {
                    "type": "integer",
                    "default": 100,
                    "description": "The maximum number of functions to return from the search per response.",
                    "minimum": 1,
                    "maximum": 1000,
                },
                "offset": {
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                    "description": "Pagination offset.",
                },
            },
            "required": [],
            "additionalProperties": False,
        },
    },
}



"""
This module defines the ACI_SEARCH_APPS meta function, which is used by LLM to search for
relevant applications (which include a set of functions) that can help complete a task.
"""

NAME = "ACI_SEARCH_APPS"

SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to find relevant apps (which includeds a set of functions) that can help complete your tasks or get data and information you need.",
        "parameters": {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Use this to find relevant apps you might need. Returned results of this function will be sorted by relevance to the intent. Examples include 'what's the top news in the stock market today', 'i want to automate outbound marketing emails'.",
                },
                "limit": {
                    "type": "integer",
                    "default": 100,
                    "description": "The maximum number of apps to return from the search.",
                    "minimum": 1,
                    "maximum": 1000,
                },
                "offset": {
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                    "description": "Pagination offset.",
                },
            },
            "required": [],
            "additionalProperties": False,
        },
    },
}



import logging
import os
import shutil

from typing_extensions import override

from aipolabs.utils._type_check import is_dict

logger: logging.Logger = logging.getLogger("aipolabs")
httpx_logger: logging.Logger = logging.getLogger("httpx")


SENSITIVE_HEADERS = {"x-api-key", "authorization"}
AIPOLABS_LOG_LEVEL = os.environ.get("AIPOLABS_LOG_LEVEL", "warn")


def setup_logging() -> None:
    logging.basicConfig(
        format="[%(asctime)s - %(name)s:%(lineno)d - %(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    httpx_logger.setLevel(logging.WARN)
    # TODO: avoid logging any sensitive headers
    logger.addFilter(SensitiveHeadersFilter())

    if AIPOLABS_LOG_LEVEL == "debug":
        logger.setLevel(logging.DEBUG)
    elif AIPOLABS_LOG_LEVEL == "info":
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARN)


class SensitiveHeadersFilter(logging.Filter):
    @override
    def filter(self, record: logging.LogRecord) -> bool:
        if is_dict(record.args):
            headers = record.args.get("headers")
            if is_dict(headers):
                headers = record.args["headers"] = {**headers}
                for header in headers:
                    if str(header).lower() in SENSITIVE_HEADERS:
                        headers[header] = "<redacted>"
        return True


def create_headline(title: str, fill_char: str = "-") -> str:
    """Create a header that fills the terminal width with the given title centered.

    Args:
        title: The text to center in the header
        fill_char: The character to use for filling (default: "-")

    Returns:
        A string containing the formatted header with green title text
    """
    # ANSI escape codes for green text and reset
    GREEN = "\033[32m"
    RESET = "\033[0m"

    terminal_width = shutil.get_terminal_size().columns or 80
    padded_title = f" {GREEN}{title}{RESET} "  # Add green color to title
    padding = fill_char * ((terminal_width - len(title) - 2) // 2)  # -2 for the spaces
    header = f"{padding}{padded_title}{padding}"

    # Add extra fill_char if the total length is off by one due to integer division
    if len(header) - len(GREEN) - len(RESET) < terminal_width:
        header += fill_char

    return header
